rcoxph=function(time,status,x,quant=.90,wt.type="exponential",niter=5,long.output=FALSE) { 
#
# This function computes a robust estimator in the Cox model. Adapted from Tadeusz Benarski's R-code
#
# References: Bednarski (1993), Scandinavian Journal of Statistics
#                    Minder and Benarski (1996), Statistics in medicine
#
#


# ========================== auxiliary functions used in the main program =======================

                  
# --------- objective function and its gradient -------------#

obj.fct=function(b,x,t,d,wt.mat,wt.vec){

hessian=function(b,x,t,d,wt.mat,wt.vec){
e <- exp(x%*%b)
q <- c(t(e)%*%wt.mat)
st <- c(e)*wt.mat
st <- t(x)%*%st

# part 1 

mat1=NULL
for(m in 1:length(b)) {
xm=c(x[,m])
lambda1=wt.vec*d/q
inter=c(lambda1)%*%sweep(t(wt.mat),2,xm*e,'*')
inter=inter%*%x
mat1=rbind(mat1,inter)
}

# part 2 

lambda2=c(wt.vec*d/q^2)
st1=sweep(st,2,sqrt(lambda2),'*')
mat2=st1%*%t(st1)

hess=mat1-mat2
hess
} 

gradient <- function(b,x,t,d,wt.mat,wt.vec){
e <- exp(x%*%b); 
q <- c(t(e)%*%wt.mat)
st <- c(e)*wt.mat ; 
st <- t(x)%*%st; 
# st[st==0]=0
-t(wt.vec*d)%*%(x-(t(st)/q))
}

    res=robust.loss(b,x,t,d,wt.mat,wt.vec)
    attr(res,"gradient")=gradient(b,x,t,d,wt.mat,wt.vec)
#    attr(res,"hessian")=hessian(b,x,t,d,wt.mat,wt.vec)
 res}


# ---------------------- weight function -------------------- #

rcoxph.wt <- function(tp,xp,bp,u,quant,wt.type){
  M=quantile(tp*exp(xp%*%bp),quant); 
  cc=tp*exp(xp%*%bp) 
  n=length(tp)

if(wt.type == "linear") wt.vec=c(M-pmin(M,cc));                  #linear weights                     
if(wt.type == "quadratic") wt.vec=c((pmax(1*M-cc,0))^2)/M^2;     #quadratic weights                    
if(wt.type == "exponential") wt.vec=c((exp(-abs(cc/M))));        #exponential weights

  ep=c(exp(xp%*%bp));

 # the loops below can be avoided  but no gain in CPU is achieved (119 sec vs 105 
 # sec). This requires the definition a lower triangular matrix of ones defined at the beginning 
 # and used as an argument all the way through (rcoxph.wt needs to use wt.matj as argument as well).
 # More CPU is need if such a matrix is recomputed at each step making the change useless!
 # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 # wt.mat=matrix(as.numeric(lower.tri(matrix(0,ncol=n,nrow=n))),ncol=n)+diag(rep(1,n))
 # if(wt.type == "exponential"){
 #    wt.mat=sweep(ep*wt.matj,2,tp,"*")
 #    fexp=function(u){abs(sign(u))*exp(-u/M)}
 #    wt.mat=apply(wt.mat,2,fexp)}



  wt.mat<-matrix(0,ncol=n,nrow=n)
  if(wt.type == "linear")
     {for (i in 1:n) {wt.mat[i:n,i]=M-pmin(tp[i]*(ep[i:n]),M)}}
  if(wt.type == "quadratic")       
     {for (i in 1:n) {wt.mat[i:n,i]=(pmax(1*M-tp[i]*ep[i:n],0))^2/M^2}}  
  if(wt.type == "exponential")
     {for (i in 1:n) {wt.mat[i:n,i]=exp(-abs(tp[i]*(ep[i:n]))/M)}}
  if (u==1) wt<-wt.vec else wt<-wt.mat; wt}

#  ----------------- robust cumulative hazard ------------------ #


robcum.hazard <- function(b,x,t,d,wt.mat,wt.vec){
#
# see reference Bednarski (1999) p. 9-10
#
n=length(t); 
p1=exp(x%*%b);
eb=t*p1;  
inter1=c(p1)%*%wt.mat
inter2 <-((d[wt.vec>0]*wt.vec[wt.vec>0])/(inter1[wt.vec>0]))
term2sum <-0*wt.vec
term2sum[wt.vec>0]<- inter2
lambda<-cumsum(c(term2sum))
lambda
}


# ----------- robust loss function (minus rob. LL)- -------------# 


robust.loss <- function(b,x,t,d,wt.mat,wt.vec){ 
e <- c(exp(x%*%b)) 
sum1 <- e%*%wt.mat
sum1[sum1==0] <-1; 
w <- (d*wt.vec)*(log(sum1)-log(e))
sum(w)
}

# NB: problem stated as a minimisation problem. we're not solving the first order 
# equation. just see that the first order equation is the derivative of a
# minimisation problem. Details about that to be given in the computation chapter.
# the objective function is minus the weighted log-likelihood (to minimise).


# ----------- asymptotic covariance matrix -------------------#

rob.covar <- function(b,x,t,d,cv,wt.mat,wt.vec){
out=NULL
n=length(t);
shift=NULL   
e=exp(x%*%b);
s0=c(t(e)%*%wt.mat);
s0[s0==0]=1
s1=t(x)%*%(c(e)*wt.mat);
score=wt.vec*d*(x-(t(s1)/s0))
  for (i in 1:n){
  inter1=c(wt.vec*e*d*wt.mat[i,])*t(s1); inter2=c(wt.vec*d*s0*wt.mat[i,]*e)*x     
    shift=rbind(shift,t(rep(1,n))%*%((inter2-inter1)/s0^2))   
 }
shifted.score=score-shift;
invcv=solve(cv)
 matv=cov(shifted.score%*%solve(cv))*n
# matq=t(shifted.score)%*%shifted.score/n
 matq=cov(shifted.score)
# matv=(invcv%*%matq%*%invcv)*n
out$matm=cv/n
out$matq=matq
out$asym.var=matv
out$score=score
out$shift=shift
out
}

 
rob.covar.test <- function(b,x,t,d,cv,wt.mat,wt.vec){
out=NULL
n=length(t)
shift=NULL   
e=exp(x%*%b)
s0=c(t(e)%*%wt.mat)
s0[s0==0]=1                       # avoids dividing by 0
s1=t(x)%*%(c(e)*wt.mat)
s1=t(s1)
score=wt.vec*d*(x-s1/s0)
ds0w <- wt.vec*d/s0
ds02w <- wt.vec*d/(s0^2)

for(i in 1:n) {
  shift.i=rep(0,length(x[1,]))
  tot1=0
  tot2=rep(0,length(x[1,]))
  for (j in 1:i){
  int1=ds0w[j]*wt.mat[i,j]            
  int2=ds02w[j]*wt.mat[i,j]*s1[j,]    
  tot1=tot1+int1                  
  tot2=tot2+int2
   }
  shift.i=e[i]*(tot1*x[i,]-tot2)
  shift=rbind(shift,shift.i) 
}
shifted.score=score-shift;
invcv=solve(cv)
matq=t(shifted.score)%*%shifted.score/n
matv=(invcv%*%matq%*%invcv)*n
out$matm=cv
out$matq=matq
out$asym.var=matv
out$score=score
out$shift=shift
out
}


# rob.covar1 <- function(b,x,t,d,cv,wt.mat,wt.vec){
# out=NULL 
# n=length(t); 
# result=NULL   
# e=exp(x%*%b); 
# q=c(t(e)%*%wt.mat); 
# q[q==0]=1
# st=t(x)%*%(c(e)*wt.mat);  
# s1=wt.vec*d*(x-(t(st)/q))
#  for (i in 1:n){
#   inter1=c(wt.vec*e*d*wt.mat[,i])*t(st); 
#  inter2=c(wt.vec*d*q*wt.mat[,i]*e)*x
#  result=rbind( result,t(rep(1,n))%*%((inter1-inter2)/q^2))                       }
# result=s1-result;
# out$matv=sqrt(diag(cov(result%*%solve(cv)))*n)
# out$matm=cv
# out$matq=cov(result)
# out
# }



# ====================================== Main program  =========================================



# initial definitions
# -------------------

out=NULL
pid=seq(1,length(time),1)
x=data.frame(x)
name=names(x)
k=length(x[1,])
xnam <- paste("x", 1:k, sep="")
(fmla <- as.formula(paste("Surv(time,status) ~ ", paste(xnam, collapse= "+"))))
names(x)=xnam  # needed for the formula (default names are x1-xp)



#
# standardisation of the covariates ( |max| per colum =1) to avoid numerical problems
# ----------------------------------------------------------------------------------

multiplier=apply(abs(x),2,max)
x=t(t(x)/multiplier)
x=data.frame(x)


# NB: internal notation 
# t used for time, d used for status, x used for the design matrix BUT
# xx is the design matrix x properly reordered (by increasing time across rows)
# same principle applies to tt, dd and pidpid

# sorting all vectors and design matrix by (increasing) time

mat=cbind(time,status,pid,x)
mat=mat[order(mat[,1],mat[,2]),] 
tt=mat[,1]; dd=mat[,2]; pidpid=mat[,3];  xx=as.matrix(mat[,-c(1,2,3)]);


#  Computation of the PLE  (classical Cox regression)
#  --------------------------------------------------


data=data.frame(cbind(time,status,x))

out.init=coxph(fmla,data=data)
estimate=out.init$coef
SE.naive=sqrt(diag(out.init$var))
out.init=coxph(fmla,data=data,robust=T)
estimate=out.init$coef
SE.LinWei=sqrt(diag(out.init$var))
z=estimate/SE.LinWei
p.value=floor((1-pchisq(z^2,1))*100000)/100000
PLE=data.frame(cbind(estimate/multiplier,SE.naive/multiplier,SE.LinWei/multiplier,z,p.value),row.names=name)


# Computation of the robust estimator (robust Cox regression) 
# -----------------------------------------------------------        

# call weight function
# tt= sorted times, xx=sorted design matrix, estimate=initial estimate (PLE), quant=quantile (=.90 default)

# initial step
# ~~~~~~~~~~~

wt.vec<-rcoxph.wt(tt,xx,estimate,1,quant,wt.type=wt.type); 
wt.mat<-rcoxph.wt(tt,xx,estimate,2,quant,wt.type=wt.type); 

out.robcox=nlm(obj.fct,p=estimate,x=xx,t=tt,d=dd,wt.mat=wt.mat,wt.vec=wt.vec,typsize=rep(1,k),hessian=F) 
out.robcox.init <- out.robcox

rob.estimate=c(out.robcox$estimate)

# Computation of the robust cumulative hazard and use of niter (3-5 usually) new iterations to get the
# final robust estimate. Each one uses the previous estimate as a starting point
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

for (i in 1:niter){
lambda=robcum.hazard(b=rob.estimate,x=xx,t=tt,d=dd,wt.mat=wt.mat,wt.vec=wt.vec)
wt.vec=rcoxph.wt(lambda,xx,rob.estimate,1,quant,wt.type=wt.type)               
wt.mat=rcoxph.wt(lambda,xx,rob.estimate,2,quant,wt.type=wt.type);               
out.robcox=nlm(obj.fct,p=rob.estimate,x=xx,t=tt,d=dd,wt.mat=wt.mat,wt.vec=wt.vec,typsize=rep(1,k),hessian=T) 
rob.estimate=c(out.robcox$estimate)
min<-out.robcox$minimum 
code <- out.robcox$code 
gradient <- out.robcox$gradient
}
 
# final computation of the cumulative hazard, weights and delta's bewteen the last 2 it.
# -------------------------------------------------------------------------------------

lambda=robcum.hazard(b=rob.estimate,x=xx,t=tt,d=dd,wt.mat=wt.mat,wt.vec=wt.vec)
wt.vec=rcoxph.wt(lambda,xx,rob.estimate,1,quant,wt.type=wt.type)            
wt.mat=rcoxph.wt(lambda,xx,rob.estimate,2,quant,wt.type=wt.type); 

# delta.wt.vec=max(abs(oldwt.vec-wt.vec))
# delta.wt.mat=max(abs(oldwt.mat-wt.mat))

    
# computation of the asymptotic variance and tabling the results
# --------------------------------------------------------------
  
hessian <-out.robcox$hessian # hessian of the robust loss function
hessian=1/2*(hessian+t(hessian)) # symmetrise the Hessian (avoids rounding problems)
eigen.val=eigen(hessian,symmetric=TRUE)$values
outvar<-rob.covar.test(b=rob.estimate,x=xx,t=tt,d=dd,cv=hessian,wt.mat=wt.mat,wt.vec=wt.vec) 
rob.SE=sqrt(diag(outvar$asym.var))  
estimate=rob.estimate/multiplier
SE=rob.SE/multiplier
z=estimate/SE
p.value=floor((1-pchisq(z^2,1))*100000)/100000
robest=data.frame(cbind(estimate,SE,z,p.value),row.names=name)



# store matrix M and Q computed at robest
# -----------------------------------------


# adjust previous hessian and matrices for rescaling

  hessian=hessian*(cbind(multiplier)%*%multiplier)
  hessian=1/2*(hessian+t(hessian))
  matq=outvar$matq*(cbind(multiplier)%*%multiplier)
  matm=hessian/length(status)
  min=out.robcox$minimum

 

#  restoring the initial order for weights, pid's, times etc
# ----------------------------------------------------------

mat=cbind(pidpid,dd,tt,wt.vec,lambda,xx)
mat=mat[order(mat[,1],mat[,2]),] 
wt=mat[,4]
initial.t=mat[,3]
pid=mat[,1]
lambda=mat[,5]
x=mat[,-c(1,2,3,4,5)]
x=t(t(x)*multiplier)


# values
# -------


out$PLE=PLE   
out$coefficients=robest
out$wt=wt  
out$xbeta.rob=c(as.matrix(x)%*%cbind(estimate)) 
out$Lambda.rob=lambda 
if(long.output) {out$matm=matm
                         out$matq=matq
                         out$wt.vec=wt.vec
                         out$wt.mat=wt.mat}
out
}